/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: grid.c,v 1.7 2014/12/09 05:10:44 zlb Exp $ */

/* This file implements the grid utils. */

#include <unistd.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "phg.h"



#define MEM_DUP(dest, src, count)				\
    phgInfo(2, "# mem copy: (%s) <=== (%s)\n", #src, #dest);	\
    if(src == NULL) {						\
	dest = NULL;						\
	phgInfo(2, "#   NULL copyed\n");			\
    } else {							\
	dest = phgCalloc((count), sizeof(src[0]));		\
	memcpy(dest, src, (count)*sizeof(src[0]));		\
	phgInfo(2, "#   %d copyed\n", (count));			\
    }



#if USE_MPI

/* --------------------------------------------------------------------------------
 *
 *
 *
 *    Set edge index
 *    
 *
 *
 * --------------------------------------------------------------------------------
 */
typedef INT EdgeVerts[2];
typedef struct EdgeVertsSort_ {
    INT v0;			/* vert */
    INT v1;			/* vert */
    INT ridx;			/* recv index */
    INT sidx;			/* uniq index */
} EdgeVertsSort;


static int
phgCompEdge(const void *p0, const void *p1)
/* compares two 'INT's (used in qsort and bsearch) */
{
    int i;
    EdgeVerts *vts0 = (EdgeVerts *) p0;
    EdgeVerts *vts1 = (EdgeVerts *) p1;

    i = (*vts0)[0] - (*vts1)[0];
    
    if (i < 0) {
	return -1;
    } else if (i > 0) {
	return 1;
    } else {
	i = (*vts0)[1] - (*vts1)[1];
	return (i < 0) ? -1 : ((i == 0) ? 0 : 1);
    }

}

static int
phgCompEdge2(const void *p0, const void *p1)
/* compares two 'INT's (used in qsort and bsearch) */
{
    int i;
    EdgeVertsSort *vts0 = (EdgeVertsSort *) p0;
    EdgeVertsSort *vts1 = (EdgeVertsSort *) p1;

    i = (*vts0).v0 - (*vts1).v0;
    
    if (i < 0) {
	return -1;
    } else if (i > 0) {
	return 1;
    } else {
	i = (*vts0).v1 - (*vts1).v1;
	return (i < 0) ? -1 : ((i == 0) ? 0 : 1);
    }

}

#define SORT_VERT(v0, v1) {			\
	int vv;					\
	if (v0 > v1) {				\
	    vv = v0; v0 = v1; v1 = vv;		\
	}					\
    }


static void
set_edge_index(GRID *g) 
/*
 * Set edge local and global index
 * 1. set local edge index
 * 2. mark remote vert then edge
 * 3. alltoall determine global index remote edge
 *
 * */
{
    INT i;
    SIMPLEX *e;

    /* ------------------------------------------------------------
     * 
     * Step 1. set local edge index
     *
     * ------------------------------------------------------------ */
    /* Note:
     *  phgUpdateEdges(g); is not safe.
     *  Since the local grid is not big, just use slow O(N^2) algorithm
     * */
    EdgeVerts *edge2vert;
    INT count, size;

    BYTE *vert_mark, *edge_mark;
    INT *edge_local, *edge_remote;
    INT n_edge_local = 0, n_edge_remote = 0;

    INT rank, nsend, nrecv, *scnts, *sdsps, *rcnts, *rdsps; 
    EdgeVerts *bsend, *brecv;
    INT nm, nn;
    MPI_Datatype type;
    EdgeVertsSort *recv_sort;
    INT remote_edge_count = 0;
    INT remote_start = 0, n_remote_all = 0;
    INT local_start = 0, n_local_all = 0;

    count = 0;
    edge2vert = phgCalloc((size = 100), sizeof(EdgeVerts));
    ForAllElements(g, e) {
	for (i = 0; i < NEdge; i++) {
	    INT v0 = e->verts[GetEdgeVertex(i, 0)];
	    INT v1 = e->verts[GetEdgeVertex(i, 1)];

	    SORT_VERT(v0, v1);

	    if (count >= size) {
		edge2vert = phgRealloc_(edge2vert,
					2*size * sizeof(EdgeVerts),
					size * sizeof(EdgeVerts));
		size *= 2;
	    }
	    edge2vert[count][0] = v0;
	    edge2vert[count][1] = v1;
	    count++;
	}
    }
    
    /* remove duplicate */
    qsort(edge2vert, count, sizeof(EdgeVerts), phgCompEdge);

    {
	INT i0 = 0;
	for (i = i0+1; i < count; i++) {
	    int cmp = phgCompEdge(edge2vert + i0,
				    edge2vert + i);
	    if (cmp < 0) {
		i0++;
		memcpy(edge2vert + i0,
		       edge2vert + i,
		       sizeof(EdgeVerts));
	    }
	}
	count = i0 + 1;
    }
    g->nedge = count;
    phgInfo(2, "nedge: %d\n", g->nedge);

    ForAllElements(g, e) {
	for (i = 0; i < NEdge; i++) {
	    INT v0 = e->verts[GetEdgeVertex(i, 0)];
	    INT v1 = e->verts[GetEdgeVertex(i, 1)];
	    EdgeVerts vts, *p;

	    SORT_VERT(v0, v1);
	    
	    vts[0] = v0;
	    vts[1] = v1;
	    p = bsearch(&vts, edge2vert,
			count, sizeof(EdgeVerts),
			phgCompEdge);
	    assert(p != NULL);

	    e->edges[i] = p - edge2vert;
	}
    }
    


    /* ------------------------------------------------------------
     * 
     * Step 2. mark remote edge
     *
     * ------------------------------------------------------------ */
    vert_mark = phgCalloc(g->nvert, sizeof(*vert_mark));
    edge_mark   = phgCalloc(count, sizeof(*edge_mark));
    edge_local  = phgCalloc(count, sizeof(*edge_local));
    edge_remote = phgCalloc(count, sizeof(*edge_remote));
    g->L2Gmap_edge = phgCalloc(count, sizeof(*g->L2Gmap_edge));
    
    for (i = 0; i < count; i++)
	g->L2Gmap_edge[i] = -1;

    ForAllElements(g, e) {
	for (i = 0; i < NFace; i++) {
	    if (e->bound_type[i] & REMOTE) {
		int v0 = e->verts[GetFaceVertex(i, 0)];
		int v1 = e->verts[GetFaceVertex(i, 1)];
		int v2 = e->verts[GetFaceVertex(i, 2)];
		vert_mark[v0] = 1;
		vert_mark[v1] = 1;
		vert_mark[v2] = 1;
	    }
	}
    }
    
    for (i = 0; i < count; i++) {
	INT v0 = edge2vert[i][0];
	INT v1 = edge2vert[i][1];
	if (vert_mark[v0] == 1
	    && vert_mark[v1] == 1) { /* both vert remote */
	    edge_remote[n_edge_remote++] = i;
	    edge_mark[i] = 1;	     /* edge remote */
	} else {
	    edge_local[n_edge_local++] = i;
	    edge_mark[i] = 0;	     /* edge local */
	}
    }

    scnts = phgCalloc(4 * g->nprocs, sizeof(*scnts));
    sdsps = scnts + g->nprocs;
    rcnts = sdsps + g->nprocs;
    rdsps = rcnts + g->nprocs;

    for (i = 0; i < count; i++) {
	if (edge_mark[i] == 1) {
	    nn = g->L2Gmap_vert[edge2vert[i][0]]
		+ g->L2Gmap_vert[edge2vert[i][1]];
	    /* nd = nn / g->nprocs; */
	    nm = nn % g->nprocs;

	    rank = nm;
	    scnts[rank]++;
	}
    }

    MPI_Alltoall(scnts, 1, MPI_INT, rcnts, 1, MPI_INT, g->comm);

    nsend = nrecv = 0;
    for (rank = 0; rank < g->nprocs; rank++) {
	sdsps[rank] = nsend;
	rdsps[rank] = nrecv;
	nsend += scnts[rank];
	nrecv += rcnts[rank];
    }
    
    bsend = phgAlloc((nsend + nrecv) * sizeof(*bsend));
    brecv = bsend + nsend;

    for (i = 0; i < count; i++) {
	if (edge_mark[i] != 0) {
	    INT V0 = g->L2Gmap_vert[edge2vert[i][0]];
	    INT V1 = g->L2Gmap_vert[edge2vert[i][1]];
	    nn = g->L2Gmap_vert[edge2vert[i][0]]
		+ g->L2Gmap_vert[edge2vert[i][1]];
	    /* nd = nn / g->nprocs; */
	    nm = nn % g->nprocs;

	    rank = nm;

	    SORT_VERT(V0, V1);

	    bsend[sdsps[rank]][0] = V0;
	    bsend[sdsps[rank]][1] = V1;
	    sdsps[rank]++;
	}
    }

    /* restore sdsps[] */
    for (rank = 0; rank < g->nprocs; rank++)
	sdsps[rank] -= scnts[rank];
    
    /* send lists */
    MPI_Type_contiguous(sizeof(*bsend), MPI_BYTE, &type);
    MPI_Type_commit(&type);
    MPI_Alltoallv(bsend, scnts, sdsps, type,
		 brecv, rcnts, rdsps, type, g->comm);

    /* process received lists */
    recv_sort = phgCalloc(nrecv, sizeof(*recv_sort));
    for (i = 0; i < nrecv; i++) {
	recv_sort[i].v0 = brecv[i][0];
	recv_sort[i].v1 = brecv[i][1];
	recv_sort[i].ridx = i;
    }

    qsort(recv_sort, nrecv, sizeof(*recv_sort), phgCompEdge2);


    /* count edge number */
    if (nrecv > 0) {
	INT i0 = 0;
	recv_sort[0].sidx = 0;
	for (i = 1; i < nrecv; i++) {
	    int cmp = phgCompEdge2(recv_sort + i - 1,
				       recv_sort + i);
	    if (cmp < 0) {
		i0++;
	    }
	    recv_sort[i].sidx = i0;
	}
	remote_edge_count = i0 + 1;
    }


    /* Reassign local edge number */
    {
	INT n_edge_locals[g->nprocs];
	MPI_Allgather(&n_edge_local, 1, PHG_MPI_INT, 
		      &n_edge_locals, 1, PHG_MPI_INT, g->comm);
	
	for (i = 0; i < g->rank; i++)
	    local_start += n_edge_locals[i];
	for (i = 0; i < g->nprocs; i++)
	    n_local_all += n_edge_locals[i];
    }

    /* Reassign remote edge number */
    {
	INT remote_edge_counts[g->nprocs];
	MPI_Allgather(&remote_edge_count, 1, PHG_MPI_INT, 
		      &remote_edge_counts, 1, PHG_MPI_INT, g->comm);
	
	for (i = 0; i < g->rank; i++)
	    remote_start += remote_edge_counts[i];
	remote_start += n_local_all;
	for (i = 0; i < g->nprocs; i++)
	    n_remote_all += remote_edge_counts[i];
    }

    for (i = 0; i < nrecv; i++) {
	INT i0 = recv_sort[i].ridx;
	brecv[i0][0] = brecv[i0][0] + brecv[i0][1]; /* sum check */
	brecv[i0][1] = recv_sort[i].sidx + remote_start;
    }

    /* send back */
    MPI_Alltoallv(brecv, rcnts, rdsps, type,
    		 bsend, scnts, sdsps, type, g->comm);

    
    for (i = 0; i < n_edge_local; i++) {
	g->L2Gmap_edge[edge_local[i]] = local_start + i;
    }

    for (i = 0; i < count; i++) {
	if (edge_mark[i] == 0) { /* local */
	}
	/* remote */
	else {
	    nn = g->L2Gmap_vert[edge2vert[i][0]]
		+ g->L2Gmap_vert[edge2vert[i][1]];
	    /* nd = nn / g->nprocs; */
	    nm = nn % g->nprocs;

	    rank = nm;
	    assert(bsend[sdsps[rank]][0] == nn); /* check sum */
	    g->L2Gmap_edge[i] = bsend[sdsps[rank]][1]; /* global index */
	    sdsps[rank]++;
	}
    }
    /* restore sdsps[] */
    for (rank = 0; rank < g->nprocs; rank++)
	sdsps[rank] -= scnts[rank];


    g->nedge_global = n_local_all + n_remote_all;
    phgInfo(2, "nedge global: %d\n", g->nedge_global);

    phgFree(scnts);
    phgFree(edge_local);
    phgFree(edge_remote);
    phgFree(vert_mark);
    phgFree(edge_mark);
    phgFree(bsend);
    phgFree(edge2vert);
    phgFree(recv_sort);
}







/* --------------------------------------------------------------------------------
 *
 *
 *   Set face index
 *
 *
 * -------------------------------------------------------------------------------- 
 * */
typedef INT FaceVerts[3];
typedef struct FaceVertsSort_ {
    INT v0;			/* vert */
    INT v1;			/* vert */
    INT v2;			/* vert */
    INT ridx;			/* recv index */
    INT sidx;			/* uniq index */
} FaceVertsSort;


static int
phgCompFace(const void *p0, const void *p1)
/* compares two 'INT's (used in qsort and bsearch) */
{
    int i;
    FaceVerts *vts0 = (FaceVerts *) p0;
    FaceVerts *vts1 = (FaceVerts *) p1;

    i = (*vts0)[0] - (*vts1)[0];
    
    if (i < 0) {
	return -1;
    } else if (i > 0) {
	return 1;
    } else {
	i = (*vts0)[1] - (*vts1)[1];
	if (i < 0) {
	    return -1;
	} else if (i > 0) {
	    return 1;
	} else {
	    i = (*vts0)[2] - (*vts1)[2];
	    return (i < 0) ? -1 : ((i == 0) ? 0 : 1);
	}
    }

}

static int
phgCompFace2(const void *p0, const void *p1)
/* compares two 'INT's (used in qsort and bsearch) */
{
    int i;
    FaceVertsSort *vts0 = (FaceVertsSort *) p0;
    FaceVertsSort *vts1 = (FaceVertsSort *) p1;

    i = (*vts0).v0 - (*vts1).v0;
    
    if (i < 0) {
	return -1;
    } else if (i > 0) {
	return 1;
    } else {
	i = (*vts0).v1 - (*vts1).v1;

	if (i < 0) {
	    return -1;
	} else if (i > 0) {
	    return 1;
	} else {
	    i = (*vts0).v2 - (*vts1).v2;
	    return (i < 0) ? -1 : ((i == 0) ? 0 : 1);
	}
    }

}

#define SORT_VERT3(v0, v1, v2) {		\
	INT vv;					\
	if (v0 > v1) {				\
	    vv = v0; v0 = v1; v1 = vv;		\
	}					\
	if (v0 > v2) {				\
	    vv = v0; v0 = v2; v2 = vv;		\
	}					\
	if (v1 > v2) {				\
	    vv = v1; v1 = v2; v2 = vv;		\
	}					\
	assert(v0 != v1);			\
	assert(v1 != v2);			\
	assert(v2 != v0);			\
    }


static void
set_face_index(GRID *g) 
/*
 * Set face local and global index
 * 1. set local face index
 * 2. mark remote vert then face
 * 3. alltoall determine global index remote face
 *
 * */
{
    INT i;
    SIMPLEX *e;

    /* ------------------------------------------------------------
     * 
     * Step 1. set local face index
     *
     * ------------------------------------------------------------ */
    /* Note:
     *  phgUpdateFaces(g); is not safe.
     *  Since the local grid is not big, just use slow O(N^2) algorithm
     * */
    FaceVerts *face2vert;
    INT count, size;

    BYTE *face_mark = NULL;
    INT *face_local = NULL, *face_remote = NULL;
    INT n_face_local = 0, n_face_remote = 0;

    INT rank, nsend, nrecv, *scnts, *sdsps, *rcnts, *rdsps;
    FaceVerts *bsend, *brecv;
    INT nm, nn;

    MPI_Datatype type;
    FaceVertsSort *recv_sort;
    INT remote_face_count = 0;

    INT remote_start = 0, n_remote_all = 0;
    INT local_start = 0, n_local_all = 0;

    count = 0;
    face2vert = phgCalloc((size = 100), sizeof(FaceVerts));
    ForAllElements(g, e) {
	for (i = 0; i < NFace; i++) {
	    INT v0 = e->verts[GetFaceVertex(i, 0)];
	    INT v1 = e->verts[GetFaceVertex(i, 1)];
	    INT v2 = e->verts[GetFaceVertex(i, 2)];

	    SORT_VERT3(v0, v1, v2);
	    
	    if (count >= size) {
		face2vert = phgRealloc_(face2vert,
					2*size * sizeof(FaceVerts),
					size * sizeof(FaceVerts));
		size *= 2;
	    }
	    face2vert[count][0] = v0;
	    face2vert[count][1] = v1;
	    face2vert[count][2] = v2;
	    count++;
	}
    }
    
    /* remove duplicate */
    qsort(face2vert, count, sizeof(FaceVerts), phgCompFace);

    {
	INT i0 = 0;
	for (i = i0+1; i < count; i++) {
	    int cmp = phgCompFace(face2vert + i0,
				    face2vert + i);
	    if (cmp < 0) {
		i0++;
		if (i != i0)
		    memcpy(face2vert + i0,
			   face2vert + i,
			   sizeof(FaceVerts));
	    }
	}
	count = i0 + 1;
    }
    g->nface = count;
    phgInfo(2, "pgrid nface: %d\n", g->nface);

    ForAllElements(g, e) {
	for (i = 0; i < NFace; i++) {
	    INT v0 = e->verts[GetFaceVertex(i, 0)];
	    INT v1 = e->verts[GetFaceVertex(i, 1)];
	    INT v2 = e->verts[GetFaceVertex(i, 2)];
	    FaceVerts vts, *p;

	    SORT_VERT3(v0, v1, v2);
	    
	    vts[0] = v0;
	    vts[1] = v1;
	    vts[2] = v2;
	    p = bsearch(&vts, face2vert,
			count, sizeof(FaceVerts),
			phgCompFace);
	    assert(p != NULL);

	    e->faces[i] = p - face2vert;
	}
    }
    


    /* ------------------------------------------------------------
     *
     * Step 2. mark remote face
     *
     * ------------------------------------------------------------ */
    face_mark   = phgCalloc(count, sizeof(*face_mark));
    face_local  = phgCalloc(count, sizeof(*face_local));
    face_remote = phgCalloc(count, sizeof(*face_remote));
    g->L2Gmap_face = phgCalloc(count, sizeof(*g->L2Gmap_face));
    
    for (i = 0; i < count; i++)
    	g->L2Gmap_face[i] = -1;

    ForAllElements(g, e) {
    	for (i = 0; i < NFace; i++) {
    	    if (e->bound_type[i] & REMOTE) {
		face_mark[e->faces[i]] = 1;	     /* face remote */
    	    }
    	}
    }
    
    for (i = 0; i < count; i++) {
    	if (face_mark[i] == 1) {
    	    face_remote[n_face_remote++] = i;
    	} else {
    	    face_local[n_face_local++] = i;
    	}
    }

    scnts = phgCalloc(4 * g->nprocs, sizeof(*scnts));
    sdsps = scnts + g->nprocs;
    rcnts = sdsps + g->nprocs;
    rdsps = rcnts + g->nprocs;

    for (i = 0; i < count; i++) {
    	if (face_mark[i] == 1) {
    	    nn = g->L2Gmap_vert[face2vert[i][0]]
    		+ g->L2Gmap_vert[face2vert[i][1]] 
		+ g->L2Gmap_vert[face2vert[i][2]];
    	    /* nd = nn / g->nprocs; */
    	    nm = nn % g->nprocs;

	    /* phgInfo(0, "remote face mark: %6d %6d %6d \n", */
	    /* 	    g->L2Gmap_vert[face2vert[i][0]], */
	    /* 	    g->L2Gmap_vert[face2vert[i][1]], */
	    /* 	    g->L2Gmap_vert[face2vert[i][2]] */
	    /* 	    ); */
    	    rank = nm;
    	    scnts[rank]++;
    	}
    }

    MPI_Alltoall(scnts, 1, MPI_INT, rcnts, 1, MPI_INT, g->comm);

    nsend = nrecv = 0;
    for (rank = 0; rank < g->nprocs; rank++) {
    	sdsps[rank] = nsend;
    	rdsps[rank] = nrecv;
    	nsend += scnts[rank];
    	nrecv += rcnts[rank];
    }
    
    bsend = phgAlloc((nsend + nrecv) * sizeof(*bsend));
    brecv = bsend + nsend;

    for (i = 0; i < count; i++) {
    	if (face_mark[i] == 1) {
    	    INT V0 = g->L2Gmap_vert[face2vert[i][0]];
    	    INT V1 = g->L2Gmap_vert[face2vert[i][1]];
    	    INT V2 = g->L2Gmap_vert[face2vert[i][2]];
    	    nn = g->L2Gmap_vert[face2vert[i][0]]
    		+ g->L2Gmap_vert[face2vert[i][1]]
		+ g->L2Gmap_vert[face2vert[i][2]];
    	    /* nd = nn / g->nprocs; */
    	    nm = nn % g->nprocs;

    	    rank = nm;

	    SORT_VERT3(V0, V1, V2);

    	    bsend[sdsps[rank]][0] = V0;
    	    bsend[sdsps[rank]][1] = V1;
    	    bsend[sdsps[rank]][2] = V2;
    	    sdsps[rank]++;
    	}
    }

    /* restore sdsps[] */
    for (rank = 0; rank < g->nprocs; rank++)
    	sdsps[rank] -= scnts[rank];
    
    /* send lists */
    MPI_Type_contiguous(sizeof(*bsend), MPI_BYTE, &type);
    MPI_Type_commit(&type);
    MPI_Alltoallv(bsend, scnts, sdsps, type,
    		 brecv, rcnts, rdsps, type, g->comm);

    /* process received lists */
    recv_sort = phgCalloc(nrecv, sizeof(*recv_sort));
    for (i = 0; i < nrecv; i++) {
    	recv_sort[i].v0 = brecv[i][0];
    	recv_sort[i].v1 = brecv[i][1];
    	recv_sort[i].v2 = brecv[i][2];
    	recv_sort[i].ridx = i;
    }

    qsort(recv_sort, nrecv, sizeof(*recv_sort), phgCompFace2);

    /* for (i = 0; i < nrecv; i++) { */
    /* 	phgInfo(0, "   remote face sort: %6d %6d %6d \n", */
    /* 		recv_sort[i].v0,  */
    /* 		recv_sort[i].v1, */
    /* 		recv_sort[i].v2); */
    /* } */

    /* count face number */
    if (nrecv > 0) {
    	INT i0 = 0;
    	recv_sort[0].sidx = 0;
    	for (i = 1; i < nrecv; i++) {
    	    int cmp = phgCompFace2(recv_sort + i - 1,
    				       recv_sort + i);
    	    if (cmp < 0) {
    		i0++;
    	    }
    	    recv_sort[i].sidx = i0;
    	}
    	remote_face_count = i0 + 1;
    }
    assert(2 * remote_face_count == nrecv); /* each remote face has two dups */
    /* phgInfo(0, "remote face count: %d %d\n", */
    /* 	    remote_face_count, nrecv); */

    /* Reassign local face number */
    {
    	INT n_face_locals[g->nprocs];
    	MPI_Allgather(&n_face_local, 1, PHG_MPI_INT,
    		      &n_face_locals, 1, PHG_MPI_INT, g->comm);
	
    	for (i = 0; i < g->rank; i++)
    	    local_start += n_face_locals[i];
    	for (i = 0; i < g->nprocs; i++)
    	    n_local_all += n_face_locals[i];
    }

    /* Reassign remote face number */
    {
    	INT remote_face_counts[g->nprocs];
    	MPI_Allgather(&remote_face_count, 1, PHG_MPI_INT,
    		      &remote_face_counts, 1, PHG_MPI_INT, g->comm);
	
    	for (i = 0; i < g->rank; i++)
    	    remote_start += remote_face_counts[i];
    	remote_start += n_local_all;
    	for (i = 0; i < g->nprocs; i++)
    	    n_remote_all += remote_face_counts[i];
    }

    for (i = 0; i < nrecv; i++) {
    	INT i0 = recv_sort[i].ridx;
    	brecv[i0][0] = brecv[i0][0] + brecv[i0][1] + brecv[i0][2]; /* sum check */
    	brecv[i0][1] = recv_sort[i].sidx + remote_start;
    }

    /* send back */
    MPI_Alltoallv(brecv, rcnts, rdsps, type,
    		 bsend, scnts, sdsps, type, g->comm);

    
    for (i = 0; i < n_face_local; i++) {
    	g->L2Gmap_face[face_local[i]] = local_start + i;
    }

    for (i = 0; i < count; i++) {
    	if (face_mark[i] == 0) { /* local */
    	}
    	/* remote */
    	else {
    	    nn = g->L2Gmap_vert[face2vert[i][0]]
    		+ g->L2Gmap_vert[face2vert[i][1]]
		+ g->L2Gmap_vert[face2vert[i][2]];
    	    /* nd = nn / g->nprocs; */
    	    nm = nn % g->nprocs;

    	    rank = nm;
    	    assert(bsend[sdsps[rank]][0] == nn); /* check sum */
    	    g->L2Gmap_face[i] = bsend[sdsps[rank]][1]; /* global index */
    	    sdsps[rank]++;
    	}
    }
    /* restore sdsps[] */
    for (rank = 0; rank < g->nprocs; rank++)
    	sdsps[rank] -= scnts[rank];


    g->nface_global = n_local_all + n_remote_all;
    phgInfo(0, "nface global: %d\n", g->nface_global);


    phgFree(scnts);
    phgFree(face_local);
    phgFree(face_remote);
    phgFree(face_mark);
    phgFree(bsend);
    phgFree(face2vert);
    phgFree(recv_sort);
}


typedef struct {
    INT		index;	/* global index */ 
    INT		lindex;	/* local index */ 
    int		rank;	/* process owning the entry */
    FLOAT       coord[3];
} B_LIST;


static B_LIST *blist;

typedef struct {
    INT		index;	/* global index */ 
    INT		lindex;	/* local index */ 
    INT         Gindex;	/* Global index */
    FLOAT       coord[3];
} B_LIST2;


static int
blist_comp_indices(const void *p0, const void *p1)
{
    B_LIST *b0 = blist + *((const INT *)p0), *b1 = blist + *((const INT *)p1);

    return b0->index - b1->index; /* compare global index */
}

#endif	/* USE_MPI */




GRID *
phgImportParallelGrid(int nvert, int nelem, 
		      int nvert_global, int nelem_global,
		      INT *L2Gmap_vert, INT *L2Gmap_elem,
		      double *coord, SIMPLE_TET *tet,
		      MPI_Comm comm
		      )
{
    int i, j;
    SIMPLEX *e;
    GRID *g = phgNewGrid(-1); 

    phgInfo(2, "nvert: %5d, nelem: %5d\n",
	    nvert, nelem);
    phgInfo(2, "nvert_global: %5d, nelem_global: %5d\n",
	    nvert_global, nelem_global);

#if USE_MPI
    /* Set MPI */
    MPI_Comm_size (comm, &g->nprocs );
    MPI_Comm_rank (comm, &g->rank );
    MPI_Comm_dup(comm, &g->comm);
#else	/* USE_MPI */
    g->nprocs = 1;
    g->rank = 0;
    g->comm = comm;
#endif	/* USE_MPI */

    /* --------------------------------------------------------------------------------
     * 
     * Only one proc
     *
     * -------------------------------------------------------------------------------- */
    if (g->nprocs == 1) {
	g->nleaf = nelem;
	g->nvert = nvert;
	g->nedge = 0;
	g->nface = 0;
	g->nelem = nelem;

	g->nvert_global = nvert;
	g->nedge_global = 0;
	g->nface_global = 0;
	g->nelem_global = nelem;

	g->nroot = nelem;
	g->ntree = nelem;

    
	g->verts = phgCalloc(g->nvert, sizeof(*g->verts));
	for (i = 0; i < g->nvert; i++) {
	    g->verts[i][0] = coord[i*3  ];
	    g->verts[i][1] = coord[i*3+1];
	    g->verts[i][2] = coord[i*3+2];
	}
#if USE_MPI
	g->L2Gmap_vert = NULL;
	g->L2Gmap_elem = NULL;
#endif	/* USE_MPI */
    

	/*
	 * Build elements
	 *  */
	g->roots = phgNewElements(nelem);
	g->elems = phgCalloc(g->nelem, sizeof(*g->elems));

	e = g->roots;
	for (i = 0; i < nelem; i++, e++) {
	    g->elems[i] = e;
	    e->index = i;

	    for (j = 0; j < NVert; j++)
		e->verts[j] = tet[i].verts[j];

	    /* Ordering */
	    {
		INT V0, V1, V2, V3,
		    v0, v1, v2, v3;

		V0 = e->verts[0]; //GlobalVertexP(g, e->verts[0]);
		V1 = e->verts[1]; //GlobalVertexP(g, e->verts[1]);
		V2 = e->verts[2]; //GlobalVertexP(g, e->verts[2]);
		V3 = e->verts[3]; //GlobalVertexP(g, e->verts[3]);
		v0 = v1 = v2 = v3 = 0;
		(V0 > V1) ? v0++ : v1++;
		(V0 > V2) ? v0++ : v2++;
		(V0 > V3) ? v0++ : v3++;
		(V1 > V2) ? v1++ : v2++;
		(V1 > V3) ? v1++ : v3++;
		(V2 > V3) ? v2++ : v3++;
		e->ordering = v0 | (v1 << 2) | (v2 << 4) | (v3 << 6);
	    }

	    /* boundary type, neigh */
	    for (j = 0; j < NFace; j++) {
		if (tet[i].bound_type[j] & INTERIOR) {
		    /* interior */
		    e->bound_type[j] = INTERIOR; 
		} else {
		    /* bdry */
		    e->bound_type[j] =
			tet[i].bound_type[j] & BDRY_MASK;
		}
	    }
	}

	/* ================================================================================
	 *
	 *  Build edge & face indices 
	 * 
	 * ================================================================================ */
	phgUpdateNeighbours(g);

	phgUpdateEdges(g);	

	phgUpdateFaces(g);	

	phgCheckConformity(g);
    }
#if USE_MPI
    /* --------------------------------------------------------------------------------
     * 
     * Multiply procs
     *
     * -------------------------------------------------------------------------------- */
    else {
	g->nleaf = nelem;
	g->nvert = nvert;
	g->nedge = -1;		/* set afterwards */
	g->nface = -1;		/* set afterwards */
	g->nelem = nelem;

	g->nvert_global = nvert_global;
	g->nedge_global = -1;	/* set afterwards */
	g->nface_global = -1;	/* set afterwards */
	g->nelem_global = nelem_global;

	g->nroot = nelem;
	g->ntree = nelem;

    
	g->verts = phgCalloc(g->nvert, sizeof(*g->verts));
	for (i = 0; i < g->nvert; i++) {
	    g->verts[i][0] = coord[i*3  ];
	    g->verts[i][1] = coord[i*3+1];
	    g->verts[i][2] = coord[i*3+2];
	}
	MEM_DUP(g->L2Gmap_vert,  L2Gmap_vert, nvert); /* Other copy */
	MEM_DUP(g->L2Gmap_elem,  L2Gmap_elem, nelem); /* Other copy */

    

	/*
	 * Build elements
	 *
	 *  */
	g->roots = phgNewElements(nelem);
	g->elems = phgCalloc(g->nelem, sizeof(*g->elems));

	e = g->roots;
	for (i = 0; i < nelem; i++, e++) {
	    g->elems[i] = e;
	    e->index = i;

	    for (j = 0; j < NVert; j++)
		e->verts[j] = tet[i].verts[j];

	    /* Ordering */
	    {
		INT V0, V1, V2, V3,
		    v0, v1, v2, v3;

		V0 = g->L2Gmap_vert[e->verts[0]]; //GlobalVertexP(g, e->verts[0]);
		V1 = g->L2Gmap_vert[e->verts[1]]; //GlobalVertexP(g, e->verts[1]);
		V2 = g->L2Gmap_vert[e->verts[2]]; //GlobalVertexP(g, e->verts[2]);
		V3 = g->L2Gmap_vert[e->verts[3]]; //GlobalVertexP(g, e->verts[3]);
		v0 = v1 = v2 = v3 = 0;
		(V0 > V1) ? v0++ : v1++;
		(V0 > V2) ? v0++ : v2++;
		(V0 > V3) ? v0++ : v3++;
		(V1 > V2) ? v1++ : v2++;
		(V1 > V3) ? v1++ : v3++;
		(V2 > V3) ? v2++ : v3++;
		e->ordering = v0 | (v1 << 2) | (v2 << 4) | (v3 << 6);
	    }

	    /* boundary type, neigh */
	    for (j = 0; j < NFace; j++) {
		if (tet[i].bound_type[j] & REMOTE) {
		    /* remote */
		    e->bound_type[j] = (INTERIOR | REMOTE);

		} else if (tet[i].bound_type[j] & INTERIOR) {
		    /* interior */
		    e->bound_type[j] = INTERIOR; 

		} else {
		    /* bdry */
		    e->bound_type[j] = tet[i].bound_type[j] & BDRY_MASK;
		    
		}
	    }

	}

	phgUpdateNeighbours(g);



	/* ================================================================================
	 *
	 *  Build edge & face indices 
	 * 
	 * ================================================================================ */

	set_edge_index(g);
	if ((g->flags & EDGE_FLAG)) {
	    phgInfo(2, "Rebuild edge flag\n");
	    /* Note: set index to global */
	    ForAllElements(g, e)
		for (i = 0; i < NEdge; i++)
		    e->edges[i] = g->L2Gmap_edge[e->edges[i]];
	    phgRebuildL2Gmap_(g, EDGE);
	}

	set_face_index(g);
	if ((g->flags & FACE_FLAG)) {
	    phgInfo(2, "Rebuild face flag\n");
	    /* Note: set index to global */
	    ForAllElements(g, e)
		for (i = 0; i < NFace; i++)
		    e->faces[i] = g->L2Gmap_face[e->faces[i]];
	    phgRebuildL2Gmap_(g, FACE);
	}

	phgInfo(2, "Update RNeigh\n");
	phgUpdateRNeighbours_(g);
    }
#endif	/* USE_MPI */


    /* ================================================================================
     *
     *  Update Boundary type
     * 
     * ================================================================================ */
    phgInfo(2, "Update bdry types\n");
    phgUpdateBoundaryTypes(g);


    /* Check */
    ForAllElements(g, e) {
	for (i = 0; i < NEdge; i++)
	    assert(e->edges[i] >= 0
		   && e->edges[i] < g->nedge);
	for (i = 0; i < NFace; i++)
	    assert(e->faces[i] >= 0
		   && e->faces[i] < g->nface);
    }


    phgGeomInit_(g, TRUE);	

    //test_poisson(g);

#if USE_MPI
    MPI_Barrier(comm);
#endif	/* USE_MPI */

    return (GRID *)g;
}


GRID *
phgGetSubGrid(GRID *g, BOOLEAN *elem_marker)
/*
 *
 * Return sub grid of g, of which the elems are marked in
 * elem_marker[g->nelem], and with the same order of ForAllElements
 * loop.
 *
 *  */
{
#if USE_MPI
    SIMPLEX *e;
    int i, j, ielem;
    INT nvert_owned, nvert_local, nvert_global, *L2Gmap_vert;
    INT nelem_global, *L2Gmap_elem;
    int *vert_owned_map, *vert_local_map, *vert_father_map, 
	*elem_owned_map;
    int *neighs;
    int nelem, *nelems, myrank, nprocs, *ranks;
    NEIGHBOUR_DATA *neigh_data;

    MPI_Group orig_group, new_group; 
    MPI_Comm new_comm; 
    double *coord;

    int *owner_rank_vert;
    INT *owner_index_vert;
    BOOLEAN *owner_flag_vert;

    INT *nvert_owned_all, vert_start = 0, elem_start = 0;

    B_LIST2 *sbuf, *rbuf, *bl;
    INT *scnts, *sdsps, *rcnts, *rdsps, nsend, nrecv;
    int rank, index;

    MPI_Datatype type;
    SIMPLE_TET *tet, *t;
    GRID *g_new ;

    /*
     * 0.0: Get whether neighs are in sub grid
     * */
    DOF *dof_mark = phgDofNew(g, DOF_DG0, 1, "sub region mark", DofNoAction);
    phgDofSetDataByValue(dof_mark, 0.);

    ielem = 0;
    ForAllElements(g, e) {
	if (elem_marker[ielem++]) {
	    dof_mark->data[e->index] = 1.;
	}
    }
    neigh_data = phgDofInitNeighbourData(dof_mark, NULL);      
    neighs = phgCalloc(g->nelem * NFace, sizeof(*neighs)); /* store flags of neigh in subgrid */
    ielem = 0;
    ForAllElements(g, e) {
	if (elem_marker[ielem++]) {
	    for (i = 0; i < NFace; i++) {
		if (e->bound_type[i] & INTERIOR) {
		    FLOAT mark = *phgDofNeighbourData(neigh_data, e, i, 0, NULL);
		    if (mark > .5)
			neighs[e->index * NFace + i] = 1;
		}
	    }
	}
    }
    phgDofReleaseNeighbourData(&neigh_data);
    phgDofFree(&dof_mark);

    /* 0.1 Create sub communicator */
    nelem = 0;
    ielem = 0;
    nelems = phgCalloc(g->nprocs, sizeof(*nelems));
    ForAllElements(g, e) {
	if (elem_marker[ielem++]) {
	    nelem++;
	}
    }
    MPI_Allgather(&nelem, 1, PHG_MPI_INT, 
		  nelems, 1, PHG_MPI_INT, g->comm);

    nprocs = 0;
    ranks = phgCalloc(g->nprocs, sizeof(*ranks));
    for (i = 0; i < g->nprocs; i++)
	if (nelems[i] > 0) {
	    ranks[nprocs++] = i;
	}


    MPI_Comm_group(g->comm, &orig_group); 
    MPI_Group_incl(orig_group, nprocs, ranks, &new_group);
    MPI_Comm_create(g->comm, new_group, &new_comm); 

    if (nelem == 0) {
	phgFree(neighs);
	phgFree(nelems);
	phgFree(ranks);
	return NULL;
    }

    MPI_Comm_rank(new_comm, &myrank); 
    phgInfo(2, "subregion rank: %d -> %d\n", g->rank, myrank);


    /* 1. build nvert, nvert_global, L2Gmap_vert.
     *    vert_local_map[g->nvert]: index of new local vert
     *    vert_owned_map[g->nvert]: index of new owned vert
     * */
    vert_owned_map = phgCalloc(g->nvert, sizeof(*vert_owned_map));
    vert_local_map = phgCalloc(g->nvert, sizeof(*vert_local_map));
    for (i = 0; i < g->nvert; i++) {
	vert_owned_map[i] = -1;
	vert_local_map[i] = -1;
    }

    nvert_local = 0;
    ielem = 0;
    ForAllElements(g, e) {
	if (elem_marker[ielem++]) {
	    for (i = 0; i < NVert; i++) {
		assert(e->verts[i] < g->nvert);
		if (vert_local_map[e->verts[i]] == -1) {
		    vert_local_map[e->verts[i]] = nvert_local++;
		}
	    }
	}
    }

    coord = phgCalloc(nvert_local * Dim, sizeof(*coord));

    ielem = 0;
    ForAllElements(g, e) {
	if (elem_marker[ielem++]) {
	    for (i = 0; i < NVert; i++) {
		if (vert_local_map[e->verts[i]] != -1) {
		    INT ivert = vert_local_map[e->verts[i]];
		    coord[3*ivert  ] = g->verts[e->verts[i]][0];
		    coord[3*ivert+1] = g->verts[e->verts[i]][1];
		    coord[3*ivert+2] = g->verts[e->verts[i]][2];
		}
	    }
	}
    }


    /*
     * Build vert ownership,
     * 
     */
    owner_rank_vert = phgAlloc(g->nvert * sizeof(*owner_rank_vert));
    owner_index_vert = phgAlloc(g->nvert * sizeof(*owner_index_vert));
    owner_flag_vert = phgCalloc(g->nvert, sizeof(*owner_flag_vert));

    for (i = 0; i < g->nvert; i++) {
	owner_rank_vert[i] = -1;
	owner_index_vert[i] = -1;
    }

    {
	/* count entries which have REMOTE bit set */
	int nsend, nrecv, *scnts, *sdsps, *rcnts, *rdsps;
	int rank, ii, jj, n, n0;
	B_LIST *bsend, *brecv, *bl = NULL;
	MPI_Datatype type;
	INT *rind;		/* sorted indices of brecv[] */
	scnts = phgCalloc(4 * nprocs, sizeof(*scnts));
	sdsps = scnts + nprocs;
	rcnts = sdsps + nprocs;
	rdsps = rcnts + nprocs;
	for (i = 0; i < g->nvert; i++) {
	    if (vert_local_map[i] != -1 &&
		g->types_vert[i] & REMOTE) {
		j = GlobalVertex(g, i);
		rank = j % nprocs;
		scnts[rank]++;
	    }
	}

	MPI_Alltoall(scnts, 1, MPI_INT, rcnts, 1, MPI_INT, new_comm);

	nsend = nrecv = 0;
	for (rank = 0; rank < nprocs; rank++) {
	    sdsps[rank] = nsend;
	    rdsps[rank] = nrecv;
	    nsend += scnts[rank];
	    nrecv += rcnts[rank];
	}
	phgInfo(3, "nsend = %d, nrecv = %d, ntotal = %d\n", nsend, nrecv,
		g->nvert);
	
	/* collect entries which have REMOTE bit */
	bsend = phgAlloc((nsend + nrecv) * sizeof(*bsend));
	brecv = bsend + nsend;

	for (i = 0; i < g->nvert; i++) {
	    if (vert_local_map[i] != -1 &&
		g->types_vert[i] & REMOTE) {
		j = GlobalVertex(g, i);
		rank = j % nprocs;
		bl = bsend + (sdsps[rank]++);

		bl->index  = j;	/* father global idx */
		bl->lindex = i;	/* father local index */
		bl->rank   = myrank; /* rank */
		bl->coord[0] = g->verts[i][0];
		bl->coord[1] = g->verts[i][1];
		bl->coord[2] = g->verts[i][2];
	    }
	}
	
	for (ii = 0; ii < nsend; ii++) {
	    bl = bsend + ii;
	    phgInfo(3, "-vert :%5d,%5d rank: %d\n", 
		    bl->lindex, bl->index, bl->rank);
	}

	/* restore sdsps[] */
	for (rank = 0; rank < nprocs; rank++)
	    sdsps[rank] -= scnts[rank];

	/* send lists */
	MPI_Type_contiguous(sizeof(*bsend), MPI_BYTE, &type);
	MPI_Type_commit(&type);
	MPI_Alltoallv(bsend, scnts, sdsps, type, brecv, rcnts, rdsps, type, new_comm);

	/* process received lists */
	rind = phgAlloc(nrecv * sizeof(*rind));
	for (ii = 0; ii < nrecv; ii++)
	    rind[ii] = ii;
	blist = brecv;
	qsort(rind, nrecv, sizeof(*rind), blist_comp_indices);
	for (ii = 0; ii < nrecv; ) {
	    for (n = ii + 1;
		 n < nrecv && blist_comp_indices(rind + ii, rind + n) == 0;
		 n++);

	    // Warning:
	    //
	    //    Min rank get ownership: Vert and edge
	    // 
	    /* the entries in the range [ii, n) correspond to the same entity */
	    rank = -1;		/* the owner process rank */
	    n0 = -1;		/* the local index in the owner process */
	    for (jj = ii; jj < n; jj++) {
		bl = brecv + rind[jj];
		if (rank == -1 || bl->rank < rank) { /* min rank */
		    rank = bl->rank;
		    n0 = bl->lindex;
		}
		if (FALSE) {
		    assert(fabs(bl->coord[0] - brecv[rind[ii]].coord[0]) < 1e-14);
		    assert(fabs(bl->coord[1] - brecv[rind[ii]].coord[1]) < 1e-14);
		    assert(fabs(bl->coord[2] - brecv[rind[ii]].coord[2]) < 1e-14);
		}
	    }
	    if (rank == -1)
		phgError(1, "rank=%d, index=%d\n",
			 bl->rank, bl->index);

	    /* Note: at this stage data in the range [ii, jj) are no longer
	     * needed, they are replaced with the owner index, owner rank and
	     * updated BTYPE value which will be sent back to incoming process */
	    for (jj = ii; jj < n; jj++) {
		bl = brecv + rind[jj];
		bl->rank = rank;	/* replace with owner's rank */
		bl->index = n0;	/* replace with owner's father local index */
	    }
	    ii = n;
	}
	phgFree(rind);

	/* send back lists */
	MPI_Alltoallv(brecv, rcnts, rdsps, type, bsend, scnts, sdsps, type, new_comm);
	MPI_Type_free(&type);

	/* process received list */
	for (ii = 0; ii < nsend; ii++) {
	    bl = bsend + ii;
	    if (bl->rank == myrank && bl->index == bl->lindex) 
		owner_flag_vert[bl->lindex] = TRUE;
	    
	    owner_rank_vert[bl->lindex] = bl->rank;
	    owner_index_vert[bl->lindex] = bl->index;
	    phgInfo(3, "+vert :%5d,%5d rank: %2d [%2d]\n", 
		    bl->lindex, bl->index, bl->rank, myrank);
	}

	phgFree(bsend);
	phgFree(scnts);
    }

    nvert_owned = 0;
    ielem = 0;
    ForAllElements(g, e) {
	if (elem_marker[ielem++]) {
	    for (i = 0; i < NVert; i++) {
		if (vert_owned_map[e->verts[i]] == -1
		    && ((!(g->types_vert[e->verts[i]] & REMOTE)) /* local */
			|| owner_flag_vert[e->verts[i]]) /* or remote owned */
		    ) {
		    vert_owned_map[e->verts[i]] = nvert_owned++;
		}
	    }
	}
    }

    for (i = 0; i < g->nvert; i++) {
	if (vert_local_map[i] != -1)
	    phgInfo(3, "   Vert %5d, %3d %3d %3d %3d (%d|%d)\n", i, 
		    vert_owned_map[i], vert_local_map[i], 
		    owner_rank_vert[i], owner_index_vert[i], 
		    owner_flag_vert[i], g->types_vert[i] & REMOTE);
    }

    nvert_owned_all = phgAlloc(nprocs * sizeof(*nvert_owned_all));
    MPI_Allgather(&nvert_owned, 1, PHG_MPI_INT, 
		  nvert_owned_all, 1, PHG_MPI_INT, new_comm);
    phgInfo(3, "owned vert: %d\n", nvert_owned_all[myrank]);

    nvert_global = 0;
    for (i = 0; i < nprocs; i++) {
	if (myrank == i)
	    vert_start = nvert_global;
	nvert_global += nvert_owned_all[i];
    }

    phgFree(nvert_owned_all);

    /* build L2G map */
    L2Gmap_vert = phgCalloc(nvert_local, sizeof(*L2Gmap_vert));
    /* owned vert */
    for (i = 0; i < g->nvert; i++) {
	if (vert_owned_map[i] != -1) {	/* owned */
	    assert(vert_local_map[i] != -1);
	    L2Gmap_vert[vert_local_map[i]] =
		vert_start + vert_owned_map[i];
	}
    }

    /* local not owned */
    scnts = phgCalloc(4 * nprocs, sizeof(*scnts));
    sdsps = scnts + nprocs;
    rcnts = sdsps + nprocs;
    rdsps = rcnts + nprocs;

    for (i = 0; i < g->nvert; i++) {
	if (vert_local_map[i] != -1 && 
	    vert_owned_map[i] == -1) {
	    rank = owner_rank_vert[i];
	    index = owner_index_vert[i];
	    phgInfo(3, "non owned %d  rank %d index %d\n", 
		    i, rank, index);
	    assert(rank >= 0);
	    scnts[rank]++;
	}
    }

    MPI_Alltoall(scnts, 1, MPI_INT, rcnts, 1, MPI_INT, new_comm);
    nsend = nrecv = 0;
    for (rank = 0; rank < nprocs; rank++) {
	sdsps[rank] = nsend;
	rdsps[rank] = nrecv;
	nsend += scnts[rank];
	nrecv += rcnts[rank];
    }

    sbuf = phgCalloc(nsend + nrecv, sizeof(*sbuf));
    rbuf = sbuf + nsend;
    for (i = 0; i < g->nvert; i++) {
	if (vert_local_map[i] != -1 && 
	    vert_owned_map[i] == -1) {
	    rank = owner_rank_vert[i];
	    index = owner_index_vert[i];
	    bl = sbuf + sdsps[rank]++;

	    bl->lindex = i; 	/* local index */
	    bl->index = index; 	/* remote index */
	    bl->Gindex = -1;
	    phgInfo(3, "Vert1: %5d(%5d) %5d %5d\n",
	    	    bl->lindex, i, bl->index, bl->Gindex);
	}
    }
    for (rank = 0; rank < nprocs; rank++)
	sdsps[rank] -= scnts[rank]; /* restore */

    /* send lists */
    MPI_Type_contiguous(sizeof(*sbuf), MPI_BYTE, &type);
    MPI_Type_commit(&type);
    MPI_Alltoallv(sbuf, scnts, sdsps, type,
		  rbuf, rcnts, rdsps, type, new_comm);

    /* local index -> global index */
    for (i = 0; i < nrecv; i++) {
    	index = rbuf[i].index;
    	// phgInfo(2, "  own %5d %d\n", index, vert_owned_map[index]);
    	assert (vert_owned_map[index] != -1); 	/* owned */
    	rbuf[i].Gindex = vert_start + vert_owned_map[index];
    	phgInfo(3, "  remote %5d -> %5d\n", index, rbuf[i].Gindex);
    }

    MPI_Alltoallv(rbuf, rcnts, rdsps, type,
		  sbuf, scnts, sdsps, type, new_comm);

    for (i = 0; i < g->nvert; i++) {
	if (vert_local_map[i] != -1 && 
	    vert_owned_map[i] == -1) {

	    rank = owner_rank_vert[i];
	    bl = sbuf + sdsps[rank]++;
	    L2Gmap_vert[vert_local_map[i]] 
		= bl->Gindex;

	    phgInfo(3, "Vert3: %5d(%5d) %5d %5d\n", 
		    bl->lindex, i, bl->index, bl->Gindex);
	}
    }


    /* Map subgrid vert to grid vert */
    vert_father_map = phgCalloc(nvert_local, sizeof(*vert_father_map));
    for (i = 0; i < g->nvert; i++)
	if (vert_local_map[i] != -1)
	    vert_father_map[vert_local_map[i]] = i;

    for (i = 0; i < nvert_local; i++) {
	assert(L2Gmap_vert[i] != -1);
	phgInfo(3, "GVERT %4d %4d (%f %f %f)\n", 
		vert_father_map[i], L2Gmap_vert[i], 
		g->verts[vert_father_map[i]][0],
		g->verts[vert_father_map[i]][1],
		g->verts[vert_father_map[i]][2]
		);
    }
    phgFree(vert_father_map);
    phgFree(sbuf);
    phgFree(scnts);





    /* 2. build nelem, nelem_global, L2Gmap_elem.
     * */
    MPI_Allgather(&nelem, 1, PHG_MPI_INT, 
		  nelems, 1, PHG_MPI_INT, new_comm);
    nelem_global = 0;
    for (i = 0; i < nprocs; i++) {
	if (myrank == i)
	    elem_start = nelem_global;
	nelem_global += nelems[i];
    }
    L2Gmap_elem = phgCalloc(nelem, sizeof(*L2Gmap_elem));
    for (i = 0; i < nelem; i++)
	L2Gmap_elem[i] = elem_start + i; 

    /* Map of elem index to subgrid elem index */
    elem_owned_map = phgCalloc(g->nelem, sizeof(*elem_owned_map));
    for (i = 0; i < g->nelem; i++)
	elem_owned_map[i] = -1;
    nelem = 0;
    ielem = 0;
    ForAllElements(g, e) {
	if (elem_marker[ielem++]) {
	    elem_owned_map[e->index] = nelem++;
	}
    }
    


    /*
     *
     * Connections
     *
     * */

    tet = phgCalloc(nelem, sizeof(*tet));
    t = tet;
    ielem = 0;
    ForAllElements(g, e) {
	if (elem_marker[ielem++]) {
	    for (i = 0; i < NVert; i++)
		t->verts[i] = vert_local_map[e->verts[i]];
	    for (i = 0; i < NFace; i++) {

		if (neighs[e->index * NFace + i]) {/* have neigh, interior */
		    if (e->bound_type[i] & REMOTE) {/* remote */
			t->bound_type[i] = (REMOTE | INTERIOR);
		    }
		    else {
			t->bound_type[i] = INTERIOR;
		    }
		} else {	/* boundary */
		    t->bound_type[i] = e->bound_type[i] & BDRY_MASK;
		}

	    }
	    t++;
	}
    }


    g_new = phgImportParallelGrid(
					nvert_local,  nelem, 
					nvert_global, nelem_global,
					L2Gmap_vert,  L2Gmap_elem,
					coord, tet,
					new_comm
					);


    phgFree(neighs);
    phgFree(nelems);
    phgFree(ranks);
    phgFree(coord);
    phgFree(tet);
    phgFree(vert_owned_map);
    phgFree(vert_local_map);
    phgFree(owner_rank_vert);
    phgFree(owner_index_vert);
    phgFree(owner_flag_vert);
    phgFree(elem_owned_map);
    phgFree(L2Gmap_vert);
    phgFree(L2Gmap_elem);

    return g_new;
#else
	SIMPLEX *e;
    int i, nvert, nelem, ielem;
    int *vert_map;
    int *neighs;
    double *coord;
    SIMPLE_TET *tet, *t;
    GRID *g_new;

    /* Get whether neighs are in sub grid */
    DOF *dof_mark = phgDofNew(g, DOF_DG0, 1, "sub region mark", DofNoAction);
    phgDofSetDataByValue(dof_mark, 0.);
    ielem = 0;
    ForAllElements(g, e) {
	if (elem_marker[ielem++]) {
	    dof_mark->data[e->index] = 1.;
	}
    }

    neighs = phgCalloc(g->nelem * NFace, sizeof(*neighs)); /* store flags of neigh in subgrid */
    ielem = 0;
    ForAllElements(g, e) {
	if (elem_marker[ielem++]) {
	    for (i = 0; i < NFace; i++) {
		if (e->bound_type[i] & INTERIOR) {
		    SIMPLEX *e_neigh = (SIMPLEX *) e->neighbours[i];
		    FLOAT mark = dof_mark->data[e_neigh->index];
		    if (mark > .5)
			neighs[e->index * NFace + i] = 1;
		}
	    }
	}
    }
    phgDofFree(&dof_mark);


    /* count subgrid vert */
    vert_map = phgCalloc(g->nvert, sizeof(*vert_map));
    for (i = 0; i < g->nvert; i++) {
	vert_map[i] = -1;
    }

    nvert = 0;
    ielem = 0;
    ForAllElements(g, e) {
	if (elem_marker[ielem++]) {
	    for (i = 0; i < NVert; i++) {
		if (vert_map[e->verts[i]] == -1) {
		    vert_map[e->verts[i]] = nvert++;
		}
	    }
	}
    }

    /* count subgrid elem */
    ielem = 0;
    nelem = 0;
    ForAllElements(g, e) {
	if (elem_marker[ielem++]) {
	    nelem++;
	}
    }

    /* copy coord */
    coord = phgCalloc(nvert * Dim, sizeof(*coord));
    ielem = 0;
    ForAllElements(g, e) {
	if (elem_marker[ielem++]) {
	    for (i = 0; i < NVert; i++) {
		if (vert_map[e->verts[i]] != -1) {
		    INT ivert = vert_map[e->verts[i]];
		    coord[3*ivert  ] = g->verts[e->verts[i]][0];
		    coord[3*ivert+1] = g->verts[e->verts[i]][1];
		    coord[3*ivert+2] = g->verts[e->verts[i]][2];
		}
	    }
	}
    }
    
    /* subgrid elem */
    tet = phgCalloc(nelem, sizeof(*tet));
    t = tet;
    ielem = 0;
    ForAllElements(g, e) {
	if (elem_marker[ielem++]) {
	    for (i = 0; i < NVert; i++)
		t->verts[i] = vert_map[e->verts[i]];

	    for (i = 0; i < NFace; i++) {
		if (neighs[e->index * NFace + i]) {/* have neigh, interior */
		    t->bound_type[i] = INTERIOR;
		} else {	/* boundary */
		    t->bound_type[i] = e->bound_type[i] & BDRY_MASK;
		}
	    }
	    t++;
	}
    }


    g_new = phgImportParallelGrid(
					nvert, nelem, 
					nvert, nelem,
					NULL,  NULL,
					coord, tet,
					0
					);
    phgFree(coord);
    phgFree(tet);
    phgFree(vert_map);
    phgFree(neighs);

    return g_new;
#endif	/* USE_MPI */
}







